﻿{*************************************************************************
                PilotLogic Software House

  Package pl_GLScene
  This file is part of CodeTyphon Studio (https://www.pilotlogic.com/)

 ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 2.0
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GLScene Project
 *
 * ***** END LICENSE BLOCK *****
 ************************************************************************}

{ GLGLXContext

   GLX specific Context.

}

Unit GLContext_GLX;

Interface

{$I GLScene.inc}

Uses
   Classes, SysUtils, LCLType,
   x, xlib, xutil,
   GLStrings,
   GLCrossPlatform, GLContext, OpenGLTokens, OpenGLAdapter,
   GLState, GLSLog;  //=== ct9999 =====

Type
   TGLXFBConfigArray = Array[0..MaxInt Div (SizeOf(GLXFBConfig) * 2)] Of GLXFBConfig;
   PGLXFBConfigArray = ^TGLXFBConfigArray;

   // TGLGLXContext

   {: A context driver for GLX. }
   TGLGLXContext = Class(TGLContext)
   Private

      FDisplay: PDisplay;
      FCurScreen: Integer;
      FDC: GLXDrawable;
      FRC: GLXContext;
      FShareContext: TGLGLXContext;
      FHPBUFFER: GLXPBuffer;
      FCurXWindow: HWND;
      FiAttribs: Packed Array Of Integer;
      FFBConfigs: PGLXFBConfigArray;
      FNewTypeContext: Boolean;
      Procedure ChooseGLXFormat;
      Function CreateTempWnd: TWindow;
      Procedure DestroyTmpWnd(AWin: TWindow);
      Procedure CreateOldContext;
      Procedure CreateNewContext;
      Procedure Validate;
      Function _glXMakeCurrent(dpy: PDisplay; draw: GLXDrawable; ctx: GLXContext): Boolean;
   Protected
      { Protected Declarations }
      Procedure ClearIAttribs;
      Procedure FreeIAttribs;
      Procedure AddIAttrib(attrib, Value: Integer);
      Procedure ChangeIAttrib(attrib, newValue: Integer);
      Procedure DropIAttrib(attrib: Integer);

      Procedure DestructionEarlyWarning(Sender: TObject);

    {: DoGetHandles must be implemented in child classes,
       and return the display + window }
      Procedure DoGetHandles(outputDevice: HWND; out XWin: HWND); Virtual; Abstract;
      Procedure GetHandles(AWindowHandle: HWND);
      Procedure DoCreateContext(ADeviceHandle: HDC); Override;
      Procedure DoCreateMemoryContext(ADeviceHandle: HWND; Width, Height: Integer; BufferCount: Integer = 1); Override;
      Function DoShareLists(aContext: TGLContext): Boolean; Override;
      Procedure DoDestroyContext; Override;
      Procedure DoActivate; Override;
      Procedure DoDeactivate; Override;

      Property DC: GLXDrawable read FDC;
      Property RenderingContext: GLXContext read FRC;
      Property CurXWindow: HWND read FCurXWindow;
   Public

      Constructor Create; Override;
      Destructor Destroy; Override;

      Function IsValid: Boolean; Override;
      Procedure SwapBuffers; Override;

      Function RenderOutputDevice: Pointer; Override;
   End;




Implementation






Resourcestring
   cForwardContextFailed = 'Can not create OpenGL 3.x Forward Context';

   // ------------------
   // ------------------ TGLGLXContext ------------------
   // ------------------

Procedure TGLGLXContext.ClearIAttribs;
Begin
   SetLength(FiAttribs, 1);
   FiAttribs[0] := 0;
End;

Procedure TGLGLXContext.FreeIAttribs;
Begin
   SetLength(FiAttribs, 0);
End;

Procedure TGLGLXContext.AddIAttrib(attrib, Value: Integer);
Var
   n: Integer;
Begin
   n := Length(FiAttribs);
   SetLength(FiAttribs, n + 2);
   FiAttribs[n - 1] := attrib;
   FiAttribs[n] := Value;
   FiAttribs[n + 1] := 0;
End;

Procedure TGLGLXContext.ChangeIAttrib(attrib, newValue: Integer);
Var
   i: Integer;
Begin
   i := 0;
   While i < Length(FiAttribs) Do
   Begin
      If FiAttribs[i] = attrib Then
      Begin
         FiAttribs[i + 1] := newValue;
         Exit;
      End;
      Inc(i, 2);
   End;
   AddIAttrib(attrib, newValue);
End;

Procedure TGLGLXContext.DropIAttrib(attrib: Integer);
Var
   i: Integer;
Begin
   i := 0;
   While i < Length(FiAttribs) Do
   Begin
      If FiAttribs[i] = attrib Then
      Begin
         Inc(i, 2);
         While i < Length(FiAttribs) Do
         Begin
            FiAttribs[i - 2] := FiAttribs[i];
            Inc(i);
         End;
         SetLength(FiAttribs, Length(FiAttribs) - 2);
         Exit;
      End;
      Inc(i, 2);
   End;
End;

// Create Temp Window And GLContext 1.1


Function TGLGLXContext.CreateTempWnd: TWindow;
Const
   Attribute: Array[0..8] Of Integer = (
      GLX_RGBA, GL_TRUE,
      GLX_RED_SIZE, 1,
      GLX_GREEN_SIZE, 1,
      GLX_BLUE_SIZE, 1,
      0);
Var
   vi: PXvisualInfo;
Begin
   // Lets create temporary window with glcontext
   Result := XCreateSimpleWindow(FDisplay, XRootWindow(FDisplay, FCurScreen), 0, 0, 1, 1, 0, // need to define some realties dimensions,
      // otherwise the context will not work
      XBlackPixel(FDisplay, FCurScreen), XWhitePixel(FDisplay, FCurScreen));
   // XMapWindow(FDisplay, win); // For the test, to see micro window
   XFlush(FDisplay); // Makes XServer execute commands
   vi := glXChooseVisual(FDisplay, FCurScreen, PGLint(@Attribute));    //=== ct9999 ============
   If vi <> nil Then
      FRC := glXCreateContext(FDisplay, vi, nil, True);
   If FRC <> nil Then
      glXMakeCurrent(FDisplay, Result, FRC);
   If vi <> nil Then
      Xfree(vi);
End;

//Free Window and GLContext


Procedure TGLGLXContext.DestroyTmpWnd(AWin: TWindow);
Begin
   If FDisplay = nil Then
      Exit;

   If FRC <> nil Then
   Begin
      glXMakeCurrent(FDisplay, 0, nil);
      glXDestroyContext(FDisplay, FRC);
      FRC := nil;
   End;

   If @AWin <> nil Then
   Begin
      XDestroyWindow(FDisplay, AWin);
      XFlush(FDisplay);
   End;
End;

Procedure TGLGLXContext.CreateOldContext;
Var
   vi: PXvisualInfo;
   shareRC: GLXContext;
Begin
   ClearIAttribs;
   If FNewTypeContext Then
   Begin
      AddIAttrib(GLX_X_RENDERABLE, GL_True);
      AddIAttrib(GLX_RENDER_TYPE, GLX_RGBA_BIT);
      AddIAttrib(GLX_DRAWABLE_TYPE, GLX_WINDOW_BIT);
   End;
   ChooseGLXFormat;

   If (ServiceContext <> nil) And (Self <> ServiceContext) Then
      shareRC := TGLGLXContext(ServiceContext).FRC
   Else
   If Assigned(FShareContext) Then
      shareRC := FShareContext.FRC
   Else
      shareRC := nil;

   Try
      If FNewTypeContext Then
      Begin
         FRC := GL.XCreateNewContext(FDisplay, FFBConfigs[0], GLX_RGBA_TYPE, shareRC, True);
         If Assigned(shareRC) Then
         Begin
            If Assigned(FRC) Then
            Begin
               FSharedContexts.Add(FShareContext);
               PropagateSharedContext;
            End Else Begin
               GLSLogger.LogWarning(strFailedToShare);
               FRC := GL.XCreateNewContext(FDisplay, FFBConfigs[0], GLX_RGBA_TYPE, nil, True);
            End;
         End;
      End Else Begin
         vi := glXChooseVisual(FDisplay, FCurScreen, @FiAttribs[0]);
         If vi = nil Then
            Raise EGLContext.Create('Failed to accept attributes');
         GLSLogger.Log('GLGLXContext: DoCreateContext->GLXFormat it is choosed');
         FRC := glXCreateContext(FDisplay, vi, shareRC, True);
         If Assigned(shareRC) Then
         Begin
            If Assigned(FRC) Then
            Begin
               If Assigned(FShareContext) Then
               Begin
                  FSharedContexts.Add(FShareContext);
                  PropagateSharedContext;
               End;
            End Else Begin
               GLSLogger.LogWarning(strFailedToShare);
               FRC := glXCreateContext(FDisplay, vi, nil, True);
            End;
         End;
         XFree(vi);
      End;

      If Not GL.X_ARB_create_context Then
      Begin
         // Down flags to backcompat

         PipelineTransformation.LoadMatricesEnabled := True;
         FGL.DebugMode := False;
      End;

      Activate;
      FGL.Initialize;
      // If we are using AntiAliasing, adjust filtering hints
      If AntiAliasing In [aa2xHQ, aa4xHQ, csa8xHQ, csa16xHQ] Then
         GLStates.MultisampleFilterHint := hintNicest
      Else
      If AntiAliasing In [aa2x, aa4x, csa8x, csa16x] Then
         GLStates.MultisampleFilterHint := hintFastest
      Else
         GLStates.MultisampleFilterHint := hintDontCare;

      GLSLogger.LogInfo('Backward compatible core context successfully created');
   Finally
      If Active Then
         Deactivate;
      XFree(FFBConfigs);
   End;
End;

Procedure TGLGLXContext.CreateNewContext;
Var
   bSuccess: Boolean;
   shareRC: GLXContext;
   fnelements: TGLInt;
   vFBConfigs: PGLXFBConfigArray;
Begin
   XSync(FDisplay, False);
   bSuccess := False;
   AddIAttrib(GLX_X_RENDERABLE,1);
   AddIAttrib(GLX_X_VISUAL_TYPE,GLX_TRUE_COLOR);
   AddIAttrib(GLX_RED_SIZE,1);
   AddIAttrib(GLX_GREEN_SIZE,1);
   AddIAttrib(GLX_BLUE_SIZE,1);
   AddIAttrib(GLX_ALPHA_SIZE, 1);
   AddIAttrib(GLX_DEPTH_SIZE, 1);

   if rcoDoubleBuffered in Options then
    begin
      AddIAttrib(GLX_DOUBLEBUFFER, 1);
    end;


   if StencilBits > 0 then
    begin
      AddIAttrib(GLX_STENCIL_SIZE, 1);
    end;

   vFBConfigs := GL.XChooseFBConfig(FDisplay, FCurScreen, @FiAttribs[0], @fnelements);

   If (ServiceContext <> nil) And (Self <> ServiceContext) Then
      shareRC := TGLGLXContext(ServiceContext).FRC
   Else
   If Assigned(FShareContext) Then
      shareRC := FShareContext.FRC
   Else
      shareRC := nil;

   Try
      ClearIAttribs;
      // Initialize forward context
      If False Then
      Begin
         If FGL.VERSION_4_2 Then
         Begin
            AddIAttrib(GLX_CONTEXT_MAJOR_VERSION_ARB, 4);
            AddIAttrib(GLX_CONTEXT_MINOR_VERSION_ARB, 2);
         End Else
         If FGL.VERSION_4_1 Then
         Begin
            AddIAttrib(GLX_CONTEXT_MAJOR_VERSION_ARB, 4);
            AddIAttrib(GLX_CONTEXT_MINOR_VERSION_ARB, 1);
         End Else
         If FGL.VERSION_4_0 Then
         Begin
            AddIAttrib(GLX_CONTEXT_MAJOR_VERSION_ARB, 4);
            AddIAttrib(GLX_CONTEXT_MINOR_VERSION_ARB, 0);
         End Else
         If FGL.VERSION_3_3 Then
         Begin
            AddIAttrib(GLX_CONTEXT_MAJOR_VERSION_ARB, 3);
            AddIAttrib(GLX_CONTEXT_MINOR_VERSION_ARB, 3);
         End Else
         If FGL.VERSION_3_2 Then
         Begin
            AddIAttrib(GLX_CONTEXT_MAJOR_VERSION_ARB, 3);
            AddIAttrib(GLX_CONTEXT_MINOR_VERSION_ARB, 2);
         End Else
         If FGL.VERSION_3_1 Then
         Begin
            AddIAttrib(GLX_CONTEXT_MAJOR_VERSION_ARB, 3);
            AddIAttrib(GLX_CONTEXT_MINOR_VERSION_ARB, 1);
         End Else
         If FGL.VERSION_3_0 Then
         Begin
            AddIAttrib(GLX_CONTEXT_MAJOR_VERSION_ARB, 3);
            AddIAttrib(GLX_CONTEXT_MINOR_VERSION_ARB, 0);
         End Else
            Abort;
         AddIAttrib(GLX_CONTEXT_FLAGS_ARB, GLX_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB);
      End;

      {$IFDEF USE_LOGGING}
//    if rcoDebug in Options then
    begin
      AddIAttrib(GLX_CONTEXT_FLAGS_ARB, GLX_CONTEXT_DEBUG_BIT_ARB);
      FGL.DebugMode := True;
    end;
      {$ENDIF}

      FRC := nil;
      If Assigned(shareRC) Then
      Begin
         FRC := FGL.XCreateContextAttribsARB(FDisplay, vFBConfigs[0], shareRC, True, @FiAttribs[0]);
         If Assigned(FRC) Then
         Begin
            If Assigned(FShareContext) Then
            Begin
               FSharedContexts.Add(FShareContext);
               PropagateSharedContext;
            End;
         End Else
            GLSLogger.LogWarning(strFailedToShare);
      End;

      If Not Assigned(FRC) Then
      Begin
         FRC := FGL.XCreateContextAttribsARB(FDisplay, vFBConfigs[0], nil, True, @FiAttribs[0]);
         If Not Assigned(FRC) Then
         Begin
            GLSLogger.LogError(cForwardContextFailed);
            Abort;
         End;
      End;

      Activate;
      FGL.Initialize;
      // If we are using AntiAliasing, adjust filtering hints
      If AntiAliasing In [aa2xHQ, aa4xHQ, csa8xHQ, csa16xHQ] Then
         GLStates.MultisampleFilterHint := hintNicest
      Else
      If AntiAliasing In [aa2x, aa4x, csa8x, csa16x] Then
         GLStates.MultisampleFilterHint := hintFastest
      Else
         GLStates.MultisampleFilterHint := hintDontCare;

      If False Then
         GLSLogger.LogInfo('Forward core context successfully created')
      Else
         GLSLogger.LogInfo('Backward compatible core context successfully created');
      bSuccess := True;
   Finally
      If Active Then
         Deactivate;
      {    GLStates.ForwardContext := GLStates.ForwardContext and bSuccess;}
      PipelineTransformation.LoadMatricesEnabled := True {not GLStates.ForwardContext};
      If Assigned(vFBConfigs) Then
         XFree(vFBConfigs);
   End;
   XSync(FDisplay, False);
End;

Procedure TGLGLXContext.Validate;
Begin
   If FRC = nil Then
      Raise EGLContext.Create('Failed to create rendering context');
   If Cardinal(FRC) = GLX_BAD_CONTEXT Then
      Raise EGLContext.Create('Bad context');
   GLSLogger.Log('GLGLXContext: RenderingContext it is created');
End;

Function TGLGLXContext._glXMakeCurrent(dpy: PDisplay; draw: GLXDrawable; ctx: GLXContext): Boolean;
Begin
   If FNewTypeContext Then
      Result := GL.XMakeContextCurrent(dpy, draw, draw, ctx)
   Else
      Result := glXMakeCurrent(dpy, draw, ctx);
End;

// ChooseGLXFormat


Procedure TGLGLXContext.ChooseGLXFormat;
Var
   vFBConfigs: PGLXFBConfigArray;
   fnelements: Integer;

   Function GetFixedAttribute(Attrib: TGLInt; Param: Integer): Integer;
   Var
      I, Res, OverRes: Integer;
   Begin
    {: Appointment of a function to look for equal or approximate values
       of attributes from the list glx.
      If you just ask all the attributes
      that the user can put it out of ignorance
      Access Violation could appear as the list will be empty. }
      Result := -1;
      OverRes := -1;
      For i := 0 To fnelements - 1 Do
      Begin
         GL.XGetFBConfigAttrib(FDisplay, vFBConfigs[i], Attrib, @Res);
         If (Res > 0) And (Res <= Param) Then
            Result := res;
         If (Res > param) And (OverRes < Res) Then
            OverRes := res;
      End;
      If (Result = -1) And (i = fnelements - 1) Then
         Result := OverRes;
   End;

   Function ChooseFBConfig: Boolean;
   Begin
      If Assigned(vFBConfigs) Then
         XFree(vFBConfigs);
      vFBConfigs := GL.XChooseFBConfig(FDisplay, FCurScreen, @FiAttribs[0], @fnelements);
      Result := Assigned(vFBConfigs);
   End;

Const
   cAAToSamples: Array[aaDefault..csa16xHQ] Of Integer =
      (0, 0, 2, 2, 4, 4, 6, 8, 16, 8, 8, 16, 16);
   cCSAAToSamples: Array[csa8x..csa16xHQ] Of Integer = (4, 8, 4, 8);
Begin
   // Temporarily create a list of available attributes
   vFBConfigs := nil;
   If Not ChooseFBConfig Then
      Raise EGLContext.Create('Failed to accept attributes');

   Try
      ColorBits := GetFixedAttribute(GLX_BUFFER_SIZE, ColorBits);
      AddIAttrib(GLX_BUFFER_SIZE, ColorBits);

      If AlphaBits > 0 Then
      Begin
         AlphaBits := GetFixedAttribute(GLX_ALPHA_SIZE, AlphaBits);
         AddIAttrib(GLX_ALPHA_SIZE, AlphaBits);
      End Else
         AddIAttrib(GLX_ALPHA_SIZE, 0);

      DepthBits := GetFixedAttribute(GLX_DEPTH_SIZE, DepthBits);
      AddIAttrib(GLX_DEPTH_SIZE, DepthBits);

      If AuxBuffers > 0 Then
      Begin
         // Even if it is 0 anyway will select something from the list FFBConfigs!
         AuxBuffers := GetFixedAttribute(GLX_AUX_BUFFERS, AuxBuffers);
         AddIAttrib(GLX_AUX_BUFFERS, AuxBuffers);
      End;

      If rcoDoubleBuffered In Options Then
      Begin
         AddIAttrib(GLX_DOUBLEBUFFER, GL_TRUE);
      End;

      //Stereo not support. See glxinfo
      If rcoStereo In Options Then
      Begin
         AddIAttrib(GLX_STEREO, GL_TRUE);
      End;

      If StencilBits > 0 Then
      Begin
         StencilBits := GetFixedAttribute(GLX_STENCIL_SIZE, StencilBits);
         AddIAttrib(GLX_STENCIL_SIZE, StencilBits);
      End;

      If AccumBits > 0 Then
         AccumBits := GetFixedAttribute(GLX_ACCUM_RED_SIZE, AccumBits Div 4) + GetFixedAttribute(GLX_ACCUM_GREEN_SIZE, AccumBits Div 4) + GetFixedAttribute(GLX_ACCUM_BLUE_SIZE, AccumBits Div 4) +
            GetFixedAttribute(GLX_ACCUM_ALPHA_SIZE, AccumBits Div 4);

      If AccumBits > 0 Then
      Begin
         AddIAttrib(GLX_ACCUM_RED_SIZE, GetFixedAttribute(GLX_ACCUM_RED_SIZE, AccumBits Div 4));
         AddIAttrib(GLX_ACCUM_GREEN_SIZE, GetFixedAttribute(GLX_ACCUM_GREEN_SIZE, AccumBits Div 4));
         AddIAttrib(GLX_ACCUM_BLUE_SIZE, GetFixedAttribute(GLX_ACCUM_BLUE_SIZE, AccumBits Div 4));
         AddIAttrib(GLX_ACCUM_ALPHA_SIZE, GetFixedAttribute(GLX_ACCUM_ALPHA_SIZE, AccumBits Div 4));
      End;

      If GL.X_ARB_multisample Then
         If AntiAliasing <> aaDefault Then
         Begin
            If AntiAliasing <> aaNone Then
            Begin
               AddIAttrib(GLX_SAMPLE_BUFFERS_ARB, GL_TRUE);
               AddIAttrib(GLX_SAMPLES_ARB, GetFixedAttribute(GLX_SAMPLES_ARB, cAAToSamples[AntiAliasing]));
               If GL.X_NV_multisample_coverage And (AntiAliasing >= csa8x) And (AntiAliasing <= csa16xHQ) Then
                  AddIAttrib(GLX_COLOR_SAMPLES_NV, GetFixedAttribute(GLX_COLOR_SAMPLES_NV, cCSAAToSamples[AntiAliasing]));
            End Else
               AddIAttrib(GLX_SAMPLE_BUFFERS_ARB, GL_FALSE);
         End;
   Finally
      If Assigned(vFBConfigs) Then
         XFree(vFBConfigs);
   End;

   FFBConfigs := GL.XChooseFBConfig(FDisplay, FCurScreen, @FiAttribs[0], @fnelements);
   If FFBConfigs = nil Then
      Raise EGLContext.Create('Failed to accept attributes');
End;

Procedure TGLGLXContext.DestructionEarlyWarning(Sender: TObject);
Begin
   DestroyContext;
End;

Procedure TGLGLXContext.GetHandles(AWindowHandle: HWND);
Begin
   DoGetHandles(AWindowHandle, FCurXWindow);
End;

// DoCreateContext


Procedure TGLGLXContext.DoCreateContext(ADeviceHandle: HDC);
Var
   tempWnd: TWindow;
Begin
   // Just in case it didn't happen already.
   If Not InitOpenGL Then
      RaiseLastOSError;

   FDisplay := XOpenDisplay(nil);

   If FDisplay = nil Then
      Raise EGLContext.Create('Failed connect to XServer');

   GLSLogger.Log('GLGLXContext: DoCreateContext->Were connected to XServer');

   FCurScreen := XDefaultScreen(FDisplay);

   tempWnd := CreateTempWnd;
   GLSLogger.Log('GLGLXContext: DoCreateContext->Is created a temporary context');
   FGL.Initialize(True);
   FNewTypeContext := GL.X_VERSION_1_2 Or GL.X_VERSION_1_3 Or GL.X_VERSION_1_4;

   DestroyTmpWnd(tempWnd);
   GLSLogger.LogInfo('Temporary rendering context destroyed');

   GetHandles(HWND(ADeviceHandle));
   FDC := CurXWindow; //FDC - TWindow

   FAcceleration := chaHardware;

   If GL.X_ARB_create_context Then
      CreateNewContext
   Else
      CreateOldContext;
   Validate;

   If (ServiceContext <> nil) And (Self <> ServiceContext) Then
   Begin
      FSharedContexts.Add(ServiceContext);
      PropagateSharedContext;
   End;
End;

// DoCreateMemoryContext


Procedure TGLGLXContext.DoCreateMemoryContext(ADeviceHandle: HWND; Width, Height: Integer; BufferCount: Integer);
Var
   TempW, TempH: Integer;
   tempWnd: TWindow;
   shareRC: GLXContext;
Begin
   // Just in case it didn't happen already.
   If Not InitOpenGL Then
      RaiseLastOSError;

   FDisplay := XOpenDisplay(nil);

   If FDisplay = nil Then
   Begin
      Raise EGLContext.Create('Failed connect to XServer');
      Exit;
   End;

   GLSLogger.Log('GLGLXContext: DoCreateContext->Were connected to XServer');
   FCurScreen := XDefaultScreen(FDisplay);

   tempWnd := CreateTempWnd;
   GLSLogger.Log('GLGLXContext: DoCreateContext->Is created a temporary context');
   FGL.Initialize(True);
   FNewTypeContext := GL.X_VERSION_1_3 Or GL.X_VERSION_1_4;
   DestroyTmpWnd(tempWnd);
   GLSLogger.LogInfo('Temporary rendering context destroyed');

   If (ServiceContext <> nil) And (Self <> ServiceContext) Then
      shareRC := TGLGLXContext(ServiceContext).FRC
   Else
   If Assigned(FShareContext) Then
      shareRC := FShareContext.FRC
   Else
      shareRC := nil;

   If FNewTypeContext Then
   Begin
      ClearIAttribs;
      AddIAttrib(GLX_X_RENDERABLE, GL_True);
      AddIAttrib(GLX_RENDER_TYPE, GLX_RGBA_BIT);
      AddIAttrib(GLX_DRAWABLE_TYPE, GLX_PBUFFER_BIT);
      ChooseGLXFormat;
      Try
         FGL.XGetFBConfigAttrib(FDisplay, FFBConfigs[0], GLX_MAX_PBUFFER_HEIGHT, @TempW);
         FGL.XGetFBConfigAttrib(FDisplay, FFBConfigs[0], GLX_MAX_PBUFFER_WIDTH, @TempH);
         If Width <= TempW Then
            TempW := Width;
         If Height <= TempH Then
            TempH := Height;
         ClearIAttribs;
         AddIAttrib(GLX_PBUFFER_WIDTH, TempW);
         AddIAttrib(GLX_PBUFFER_HEIGHT, TempH);
         AddIAttrib(GLX_LARGEST_BUFFER, 0);
         FHPBUFFER := FGL.XCreatePbuffer(FDisplay, FFBConfigs[0], @FiAttribs[0]);
         If FHPBUFFER = 0 Then
            Raise EPBuffer.Create('Unabled to create pbuffer.');
         FDC := FHPBUFFER;
         GLSLogger.Log('GLGLXContext: DoCreateContext->PBuffer is Created');

         FAcceleration := chaHardware;

         FRC := FGL.XCreateNewContext(FDisplay, FFBConfigs[0], GLX_RGBA_TYPE, shareRC, True);
         If Assigned(shareRC) Then
         Begin
            If Assigned(FRC) Then
            Begin
               FSharedContexts.Add(FShareContext);
               PropagateSharedContext;
            End Else Begin
               GLSLogger.LogWarning(strFailedToShare);
               FRC := FGL.XCreateNewContext(FDisplay, FFBConfigs[0], GLX_RGBA_TYPE, nil, True);
            End;
         End;
         Validate;

         If Not FGL.X_ARB_create_context Then
         Begin
            // Down flags to backcompat
            //  GLStates.ForwardContext := False;
            PipelineTransformation.LoadMatricesEnabled := True;
            FGL.DebugMode := False;
         End;

         Activate;
         FGL.Initialize;
         // If we are using AntiAliasing, adjust filtering hints
         If AntiAliasing In [aa2xHQ, aa4xHQ, csa8xHQ, csa16xHQ] Then
            GLStates.MultisampleFilterHint := hintNicest
         Else
         If AntiAliasing In [aa2x, aa4x, csa8x, csa16x] Then
            GLStates.MultisampleFilterHint := hintFastest
         Else
            GLStates.MultisampleFilterHint := hintDontCare;

         If BufferCount > 1 Then
            FGL.DrawBuffers(BufferCount, @MRT_BUFFERS);

         If (ServiceContext <> nil) And (Self <> ServiceContext) Then
         Begin
            FSharedContexts.Add(ServiceContext);
            PropagateSharedContext;
         End;

         GLSLogger.LogInfo('Backward compatible core PBuffer context successfully created');
      Finally
         If Active Then
            Deactivate;
         XFree(FFBConfigs);
      End;
   End Else
      Raise
      EGLContext.Create('For PBuffer Context required GLX above 1.2');
End;

// DoShareLists


Function TGLGLXContext.DoShareLists(aContext: TGLContext): Boolean;
Var
   otherRC: GLXContext;
Begin
   Result := False;
   If aContext Is TGLGLXContext Then
   Begin
      otherRC := TGLGLXContext(aContext).RenderingContext;
      If RenderingContext <> nil Then
      Begin
         If (RenderingContext <> otherRC) Then
         Begin
            DestroyContext;
            FShareContext := TGLGLXContext(aContext);
            Result := True;
         End;
      End Else Begin
         FShareContext := TGLGLXContext(aContext);
         Result := False;
      End;
   End Else
      Raise Exception.Create(cIncompatibleContexts);
End;

Procedure TGLGLXContext.DoDestroyContext;
Begin
   GLSLogger.Log('GLGLXContext: DoDestroyContext');
   If Not Assigned(FDisplay) Then
      Raise EGLContext.Create('Lost connection XServer');
   If (glXGetCurrentContext() = FRC) And (Not _glXMakeCurrent(FDisplay, 0, nil)) Then
      Raise EGLContext.Create(cContextDeactivationFailed);
   glXDestroyContext(FDisplay, FRC);
   GLSLogger.Log('GLGLXContext: DoDestroyContext->RenderingContext it is Destroyd');
   If FHPBUFFER <> 0 Then
   Begin
      GL.XDestroyPbuffer(FDisplay, FHPBUFFER);
      FHPBUFFER := 0;
      GLSLogger.Log('GLGLXContext: DoDestroyContext->RenderingContext it is Destroyd');
   End;
   FRC := nil;
   FDC := 0;
   FShareContext := nil;
   XCloseDisplay(FDisplay);
   FCurScreen := 0;
End;

// DoActivate


Procedure TGLGLXContext.DoActivate;
Begin
   If Not _glXMakeCurrent(FDisplay, FDC, FRC) Then
      Raise EGLContext.Create(cContextActivationFailed);
   XSync(FDisplay, False);

   If Not FGL.IsInitialized Then
      FGL.Initialize;
End;

// Deactivate


Procedure TGLGLXContext.DoDeactivate;
Begin
   If Not _glXMakeCurrent(FDisplay, 0, nil) Then
      Raise EGLContext.Create(cContextDeactivationFailed);
End;

Constructor TGLGLXContext.Create;
Begin
   Inherited Create;
   ClearIAttribs;
End;

Destructor TGLGLXContext.Destroy;
Begin
   Inherited Destroy;
End;

// IsValid


Function TGLGLXContext.IsValid: Boolean;
Begin
   Result := (FRC <> nil);
End;

// SwapBuffers


Procedure TGLGLXContext.SwapBuffers;
Begin
   If (FDC <> 0) And (rcoDoubleBuffered In Options) Then
      glXSwapBuffers(FDisplay, FDC);
End;

Function TGLGLXContext.RenderOutputDevice: Pointer;
Begin
   Result := nil;
End;

End.
